module FNC where

import Debug.Trace (trace)

type Prop = Int

data Binop = Conj | Disj | Impl deriving Show
pp :: Binop -> String
pp Conj = " ^ "
pp Disj = " v "
pp Impl = " -> "

opposite :: Binop -> Binop
opposite Conj = Disj
opposite Disj = Conj

data Unop = Neg deriving Show

data Formula = Atom Prop
             | BO Binop Formula Formula
             | UO Unop  Formula 
               deriving Show

non = UO Neg
implique = BO Impl
impl = implique
et = BO Conj
ou = BO Disj
biimplique phi psi = (phi `implique` psi) `et` (psi `implique` phi)
ssi = biimplique
vrai = (Atom 0) `ou` (non (Atom 0))
faux = (Atom 0) `et` (non (Atom 0))



prettyPrint (Atom x) = "p_" ++ show x 
prettyPrint (BO op phi psi) = "(" ++ prettyPrint phi ++ opstr ++ prettyPrint psi ++ ")"
  where
    opstr = pp op
prettyPrint (UO _ phi) = "~" ++ prettyPrint phi

    
elImplStep :: Formula -> Either Formula Formula

elImplStep f@(Atom x) = Left f

elImplStep f@(BO Impl phi psi) =
  case elImplStep phi of
    Right phi' -> Right (BO Impl phi' psi)
    Left _ -> case elImplStep psi of
      Right psi' -> Right (BO Impl phi psi')
      Left _ ->  Right (BO Disj (UO Neg phi) psi)

elImplStep f@(BO op phi psi) =
  case elImplStep phi of
    Right phi' -> Right (BO op phi' psi)
    Left _ -> case elImplStep psi of
      Right psi' -> Right (BO op phi psi')
      Left _ -> Left f

elImplStep f@(UO Neg phi) = case elImplStep phi of
  Right phi' -> Right (UO Neg phi')
  Left _ -> Left f 



pdNeg :: Formula -> Either Formula Formula
pdNeg f@(Atom _) = Left f
pdNeg f@(UO Neg (Atom _)) = Left f
pdNeg f@(BO op phi psi) = case pdNeg phi of
    Right phi' -> Right (BO op phi' psi)
    Left _ -> case pdNeg psi of
      Right psi' -> Right (BO op phi psi')
      Left _ -> Left f

pdNeg f@(UO Neg (UO Neg phi)) =
  case pdNeg phi of
    Right phi' -> Right (UO Neg (UO Neg phi'))
    Left _ -> Right phi

pdNeg f@(UO Neg (BO op phi psi)) =
  case pdNeg phi of
    Right phi' -> Right (UO Neg (BO op phi' psi))
    Left _ -> case pdNeg psi of
      Right psi' -> Right (UO Neg (BO op phi psi'))
      Left _ -> Right ((BO (opposite op) (UO Neg phi) (UO Neg psi)))

binDo :: (a -> Either a a) ->
         (a -> a -> a) ->
         a -> a -> Either a a
binDo f constr x y = case f x of
  Right x' -> Right (constr x' y)
  Left _ -> case f y of
    Right y' ->  Right (constr x y')
    Left _ -> Left (constr x y)


distrStep f@(Atom _) = Left f
distrStep f@(UO Neg _) = Left f
distrStep (BO Conj phi psi) = binDo distrStep (BO Conj) phi psi
distrStep (BO Disj phi@(BO Conj phi1 phi2) psi) =
  case distrStep phi of
    Right phi' -> Right (BO Disj phi' psi)
    Left _ -> case distrStep psi of
      Right psi' -> Right (BO Disj phi psi')
      Left _ -> Right (BO Conj (BO Disj phi1 psi) (BO Disj phi2 psi))
distrStep (BO Disj phi psi@(BO Conj psi1 psi2)) =
  case distrStep phi of
    Right phi' -> Right (BO Disj phi' psi)
    Left _ -> case distrStep psi of
      Right psi' -> Right (BO Disj phi psi')
      Left _ -> Right (BO Conj (BO Disj phi psi1) (BO Disj phi psi2))
distrStep (BO Disj phi psi) = binDo distrStep (BO Disj) phi psi


loopStep :: (Formula -> String) ->
            (Formula -> Either Formula Formula) ->
            Formula -> Either Formula Formula
loopStep howToShow stepfunction f = trace (howToShow f)
  (stepfunction f >>= loopStep howToShow stepfunction )


eliminationImplication = trace "\n1- Éliminer l'implication :"
  loopStep prettyPrint elImplStep
  
pushdownNegation = trace "\n2- Pousser la négation vers les symboles propositionnels :"
  loopStep prettyPrint pdNeg 

distribute = trace "\nPousser les disjonctions vers les littéraux :"
  loopStep prettyPrint distrStep 

procedure :: Formula -> IO ()
procedure f0 = do
  putStrLn $ "Formule en entrée : "++ prettyPrint f0
  case eliminationImplication f0 of
    Left f1 -> case pushdownNegation f1 of
      Left f2 -> case distribute f2 of
        Left _ -> putStr "\nFin de la procedure\n"

      

